/*
 * keyb4x4.c
 *
 *  Created on: 14.6.2011
 *      Author: emlin
 */

# include "keyb4x4-ioctl.h"

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/kernel.h>

#include <asm/uaccess.h>

#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/delay.h>

// -----------------------------------------------------------------------------
#define DEV_NAME					"keyb4x4"
#define DEV_COUNT					1

#define PAR_PORT_BASE				0x378
#define PAR_PORT_SIZE				3

#define PAR_PORT_REG_DATA			(0x00)
#define PAR_PORT_REG_STAT			(0x01)
#define PAR_PORT_REG_CTRL			(0x02)

// -----------------------------------------------------------------------------
// Hodnoty signalu /SEL, /AF a /STR jsou HW invertovane
// zapne zapis na radky klavesnice
// /SEL = 0, /INT = 0, /AF = 1, /STR = 0
#define KEYB4X4_WRITE_ROW			0x02

// zapne cteni ze sloupcu klavesnice
// /EnBiDir = 1 /enIRQ = 0  /SEL = 0, /INT = 0, /AF = 1, /STR = 1
#define	KEYB4X4_READ_COL			0x23

// deaktivuje klavesnici
// /SEL = 1, /INT = 0, /AF = 1, /STR = 1
#define	KEYB4X4_DISABLE				0x0B

// -----------------------------------------------------------------------------
#define	MAX_ROW_CNT					4

#define INIT_ROW_CODE				0x01
#define MAX_ROW_CODE				(INIT_ROW_CODE << MAX_ROW_CNT)

#define	KEY_READ_CNT_TO_VALIDATE	5
#define NO_KEY_PRESSED				0x00
// msecs
#define SCAN_KEYS_PERIOD			20
//scan codes
#define SCAN_BUFFER_SIZE			16

// -----------------------------------------------------------------------------
typedef unsigned char scan_code_t;

// -----------------------------------------------------------------------------
enum
{
	ModuleExit = 0,

	Failed_ChrdevRegion,
	Failed_CdevAlloc,
	Failed_CdevAdd,
	Failed_SysClass,
	Failed_SysDev,
	Failed_ReqRegion,
	Failed_IoMap,
};

// -----------------------------------------------------------------------------
// parametry ovladace
struct keyb4x4_t
{
	// major-minor par
	dev_t				dev_num;
	// ukazatel na strukturu cdev reprezentujici znakove zarizeni
	struct cdev *		cdev_ptr;
	// ukazatel na tridu zarizeni v adresari /sys/class
	struct class *		sys_class_ptr;
	struct device * 	sys_dev_ptr;
	// hlida pocet volani operace open()
	atomic_t			open_cnt;
	// zacatek bloku registru paralelniho portu namapovaneho jako I/O pamet
	void __iomem *		port_base;
	// casovac
	struct timer_list	scan_timer;
	// promenna pro deaktivaci casovace
	// (1 = deaktivuj casovac, 0 = casovac byl aktivovan
	atomic_t 			scan_timer_stop;
	// bufer scan kodu
	unsigned char 		scan_buffer[SCAN_BUFFER_SIZE];
	// spinlock pro pristup ke sdilene promenne
	spinlock_t 			scan_buffer_lock;
	// pozice pro cteni ze scan buferu
	int 				buf_read_idx;
	// pozice pro zapis do scan buferu
	int 				buf_write_idx;
	// posledni sejmuty scan kod
	scan_code_t 		scan_code_mem;
	// pocet snimani stejne klavesy / scan kodu
	int 				key_read_cnt;
};

static struct keyb4x4_t keyb =
{
	.dev_num = MKDEV(0, 0),
	.cdev_ptr = NULL,
	.sys_class_ptr = NULL,
	.sys_dev_ptr = NULL,
	.open_cnt = ATOMIC_INIT(0),
	.port_base = NULL,
	.scan_timer_stop = ATOMIC_INIT(1),
	.scan_buffer_lock = SPIN_LOCK_UNLOCKED,
	.buf_read_idx = 0,
	.buf_write_idx = 0,
	.scan_code_mem = NO_KEY_PRESSED,
	.key_read_cnt = 0,
};

// -----------------------------------------------------------------------------
static inline void write_row_neg(unsigned char value)
{
	iowrite8(KEYB4X4_WRITE_ROW, (keyb.port_base + PAR_PORT_REG_CTRL));
	wmb();
	iowrite8(~value, (keyb.port_base + PAR_PORT_REG_DATA));
	wmb();
	iowrite8(KEYB4X4_DISABLE, (keyb.port_base + PAR_PORT_REG_CTRL));
}

// -----------------------------------------------------------------------------
static inline unsigned char read_col_neg(void)
{
	unsigned int value = 0x00;

	iowrite8(KEYB4X4_READ_COL, (keyb.port_base + PAR_PORT_REG_CTRL));
	mb();
	value = ioread8((keyb.port_base + PAR_PORT_REG_DATA));
	mb();
	iowrite8(KEYB4X4_DISABLE, (keyb.port_base + PAR_PORT_REG_CTRL));

	return (~((unsigned char) value));
}

// -----------------------------------------------------------------------------
static void clear_buffer(void)
{
	spin_lock_bh(&keyb.scan_buffer_lock);

	keyb.buf_read_idx = 0;
	keyb.buf_write_idx = 0;

	spin_unlock_bh(&keyb.scan_buffer_lock);
}

// -----------------------------------------------------------------------------
static scan_code_t read_buffer(void)
{
	scan_code_t scan_code = NO_KEY_PRESSED;

	spin_lock_bh(&keyb.scan_buffer_lock);

	// kdyz bufer neni prazdny
	if ( keyb.buf_read_idx != keyb.buf_write_idx )
	{
		// bufer je kruhovy, z konce skoc na zacatek
		scan_code = keyb.scan_buffer[keyb.buf_read_idx++];
		if ( keyb.buf_read_idx == SCAN_BUFFER_SIZE )
		{
			keyb.buf_read_idx = 0;
		}
	}

	spin_unlock_bh(&keyb.scan_buffer_lock);

	return scan_code;
}

// -----------------------------------------------------------------------------
static void write_buffer(scan_code_t scan_code)
{
	spin_lock_bh(&keyb.scan_buffer_lock);

	// bufer je kruhovy, z konce skoc na zacatek
	keyb.scan_buffer[keyb.buf_write_idx++] = scan_code;
	if ( keyb.buf_write_idx == SCAN_BUFFER_SIZE )
	{
		keyb.buf_write_idx = 0;
	}
	// index pro cteni drzim o jednu pozici vyse nez je index pro zapis
	// pokud by se totiz oba indexy rovnaly, znamenalo by to,
	// ze je bufer prazdny
	if ( keyb.buf_write_idx == keyb.buf_read_idx )
	{
		keyb.buf_read_idx++;
		if ( keyb.buf_read_idx == SCAN_BUFFER_SIZE )
		{
			keyb.buf_read_idx = 0;
		}
	}

	spin_unlock_bh(&keyb.scan_buffer_lock);
}

// -----------------------------------------------------------------------------
static void scantimer_on(void)
{
	// vynulujeme si deaktivacni promenou
	atomic_set(&keyb.scan_timer_stop, 0);

	// nastavime si znacku, kdy ma casovac vyprset
	keyb.scan_timer.expires = jiffies + msecs_to_jiffies(SCAN_KEYS_PERIOD);

	// spustime casovac
	add_timer(&keyb.scan_timer);
}

// -----------------------------------------------------------------------------
static void scantimer_off(void)
{
	atomic_set(&keyb.scan_timer_stop, 1);

	del_timer(&keyb.scan_timer);
}

// -----------------------------------------------------------------------------
static inline scan_code_t get_scan_code(unsigned char row_code,
									  	unsigned char col_code)
{
	return ( (row_code << 4) + col_code );
}

// -----------------------------------------------------------------------------
void scantimer_userfunction(unsigned long data)
{
	unsigned char row_code = INIT_ROW_CODE, col_code = NO_KEY_PRESSED;
	int key_pressed = 0, keyb_scan_done = 0;
	scan_code_t scan_code = NO_KEY_PRESSED;


	// oskenuj klavesnici zda je stisknuta nejaka klavesa
	while ( !key_pressed && !keyb_scan_done )
	{
		// aktivuj vybrany radek
		write_row_neg(row_code);
		// pockej kvuli kapacitam membranove klavesnice
		udelay(50);
		// vycti si stav sloupcu
		col_code = read_col_neg();
		// je nejaka klavesa stisknuta?
		// ano, skonci s aktivaci radku a vyhodnot klavesu
		if ( col_code != NO_KEY_PRESSED )
		{
			key_pressed = 1;

			scan_code = get_scan_code(row_code, col_code);
		}
		// neni, posun kod radku na dalsi radek
		else
		{
			row_code <<= 1;
			if ( row_code >= MAX_ROW_CODE )
			{
				row_code = INIT_ROW_CODE;
				keyb_scan_done = 1;
			}
		}
	}

	// vyhodnot stisknutou klavesnici
	if ( key_pressed )
	{
		++keyb.key_read_cnt;

		// je stisknuta klavesa stejna jako ta predchozi?
		// ano: 3x stejna klavesa => zakmity vyfiltrovane => uloz si
		//      scan kod klavesy a zacni skenovat novy stisk
		if ( scan_code == keyb.scan_code_mem )
		{
			if ( keyb.key_read_cnt >= KEY_READ_CNT_TO_VALIDATE )
			{
				write_buffer(scan_code);
				pr_alert("ScanCode: %x\n", scan_code);

				keyb.scan_code_mem = NO_KEY_PRESSED;
				keyb.key_read_cnt = 0;
			}
		}
		// ne: tato klavesa stisknuta poprve, zapamatuj si kod sloupce
		else
		{
			keyb.scan_code_mem = scan_code;
		}
	}

	// naplanujeme nove spusteni casovace
	if ( !atomic_read(&keyb.scan_timer_stop) )
	{
		mod_timer(&keyb.scan_timer,
				  jiffies + msecs_to_jiffies(SCAN_KEYS_PERIOD));
	}
}

// -----------------------------------------------------------------------------
static int keyb4x4_open(struct inode * inode_ptr, struct file * file_ptr)
{
	int result = 0, tmp_open_cnt = 0;

	// soubor zarizeni muze byt otevren jen jednim programem soucasne
	// atomicky inkrementuje promenou o 1 a vrati PUVODNI hodnotu
	tmp_open_cnt = atomic_inc_and_test(&keyb.open_cnt);
	if ( tmp_open_cnt > 1 )
	{
		pr_err("keyb4x4_open(): already opened!\n");
		atomic_dec(&keyb.open_cnt);
		result = -EBUSY;
		goto End;
	}

	// modul nepodporuje systemove volani lseek()
	result = nonseekable_open(inode_ptr, file_ptr);
	if ( result < 0 )
	{
		goto End;
	}

End:
	pr_devel("keyb4x4_open(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static int keyb4x4_release(struct inode * inode_ptr, struct file * file_ptr)
{
	// povolime nove otevreni
	atomic_dec(&keyb.open_cnt);

	pr_devel("keyb4x4_release(): done\n");

	return 0;
}

// -----------------------------------------------------------------------------
static ssize_t keyb4x4_read(struct file * file_ptr,
				   		  char __user * data_ptr, size_t data_size,
				   		  loff_t * posPtr)
{
	ssize_t result = 0;

	if ( data_size > 0 )
	{
		// vyber scan kod nejdrive stisknute klavesy
		scan_code_t scan_code = read_buffer();
		if ( scan_code != NO_KEY_PRESSED )
		{
			pr_devel("R: %x\n", scan_code);

			// zapis jej do uzivatelskeho prostoru
			// a kdyz je to ok, tak vrat velikost promenne v bytech
			result = put_user(scan_code, data_ptr);
			if ( result == 0 )
			{
				result = sizeof(scan_code);
			}
		}
	}

	pr_devel("keyb4x4_read(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static ssize_t keyb4x4_write(struct file * file_ptr,
				    const char __user * data_ptr, size_t data_size,
				    loff_t * posPtr)
{
	pr_devel("keyb4x4_write(): failed\n");

	return -EINVAL; // nelze cist;
}

// -----------------------------------------------------------------------------
static int keyb4x4_ioctl(struct inode * inode_ptr, struct file * file_ptr,
				       unsigned int cmd, unsigned long arg)
{
	int result = 0;

	// kontrola kodu prikazu
	if ( (_IOC_TYPE(cmd) != KEYB4X4_MAGIC_NO) ||
		 (_IOC_NR(cmd) > KEYB4X4_MAX_CMD_NR) )
	{
		result = -ENOTTY;
		goto End;
	}

	// dekodujeme prikaz a pak jej provedeme
	switch( cmd )
	{
		case KEYB4X4_CLEAR_BUF:
			clear_buffer();
			result = 0;
			break;

		default:
			result = -ENOTTY;
			break;
	}

End:
	pr_devel("keyb4x4_ioctl(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
// virtualni tabulka - tabulka operaci podporovanych ovladacem
static struct file_operations keyb4x4_file_ops =
{
	.owner = THIS_MODULE,
	.open = &keyb4x4_open,
	.release = &keyb4x4_release,
	.read = &keyb4x4_read,
	.write = &keyb4x4_write,
	.ioctl = keyb4x4_ioctl,
	// modul nepodporuje metodu llseek
	.llseek = &no_llseek,
};

// -----------------------------------------------------------------------------
static void cleanup(int error)
{
	switch ( error )
	{
		// kompletni uklid pri exitu
		case ModuleExit:
			// zrusime adresar modulu v /sys
			device_destroy(keyb.sys_class_ptr, keyb.dev_num);

		case Failed_SysDev:
			// zrusime tridu
			class_destroy(keyb.sys_class_ptr);

		case Failed_SysClass:
			// vyjmeme modul z jadra
			// pozn.: cdev_del() se vola jen po uspesnem cdev_add()
			cdev_del(keyb.cdev_ptr);

		case Failed_CdevAdd:
			// vypneme casovac (pokud nebezel nic se nestane)
			scantimer_off();
			// zrusime mapovani portu
			ioport_unmap(keyb.port_base);

		case Failed_IoMap:
			// zrusime rezervaci I/O portu
			release_region(PAR_PORT_BASE, PAR_PORT_SIZE);

		case Failed_ReqRegion:
			// pozn.: neexistuje opak fce cdev_alloc(), uvolneni pameti resi
			// jadro samo
		case Failed_CdevAlloc:
			// uvolnime nami zabrany major-minor par
			unregister_chrdev_region(keyb.dev_num, DEV_COUNT);

		case Failed_ChrdevRegion:
			// jeste nic se nealokovalo, takze neni co vracet
			break;

		default:
			break;
	}
}

// -----------------------------------------------------------------------------
static int __init keyb4x4_init(void)
{
	int result = 0;
	struct resource * region_ptr;

	// registrace k major-minor paru
	result = alloc_chrdev_region(&keyb.dev_num, MINOR(keyb.dev_num),
								 DEV_COUNT, DEV_NAME);
	if ( result < 0 )
	{
		cleanup(Failed_ChrdevRegion);
		goto End;
	}

	// alokace pameti pro strukturu cdev
	keyb.cdev_ptr = cdev_alloc();
	if ( keyb.cdev_ptr == NULL )
	{
		result = -ENOMEM;
		cleanup(Failed_CdevAlloc);
		goto End;
	}

	// nastaveni vlastnika a seznamu souborovych operaci
	keyb.cdev_ptr->owner = THIS_MODULE;
	keyb.cdev_ptr->ops = &keyb4x4_file_ops;

	// rezervujeme si paralelni port
	region_ptr = request_region(PAR_PORT_BASE, PAR_PORT_SIZE, DEV_NAME);
	if ( region_ptr == NULL )
	{
		pr_err("keyb4x4_init(): parallel port already in use!\n");
		cleanup(Failed_ReqRegion);
		result = -EBUSY;
		goto End;
	}

	// premapujem port jako I/O pamet
	keyb.port_base = ioport_map(PAR_PORT_BASE, PAR_PORT_SIZE);
	if ( keyb.port_base == NULL )
	{
		cleanup(Failed_IoMap);
		result = -ENOMEM;
		goto End;
	}

	// inicializace casovace (zbytek se nastavuje pri spusteni)
	keyb.scan_timer.function = &scantimer_userfunction;
	keyb.scan_timer.data = (unsigned long) NULL;
	init_timer(&keyb.scan_timer);

	// spustime casovac, tj. zacneme obsluhovat klavesnici
	scantimer_on();

//  --- pridani ovladace do jadra ---
	result = cdev_add(keyb.cdev_ptr, keyb.dev_num, DEV_COUNT);
	if ( result < 0 )
	{
		cleanup(Failed_CdevAdd);
		goto End;
	}

	// vytvorime si tridu v /sys/class
	keyb.sys_class_ptr = class_create(THIS_MODULE, DEV_NAME);
	if ( IS_ERR(keyb.sys_class_ptr) )
	{
		cleanup(Failed_SysClass);
		result = PTR_ERR(keyb.sys_class_ptr);
		goto End;
	}

	// pridame nas modul do /sys/class/keyb4x4
	keyb.sys_dev_ptr = device_create(keyb.sys_class_ptr,
									 NULL, keyb.dev_num, NULL, DEV_NAME);
	if ( IS_ERR(keyb.sys_dev_ptr) )
	{
		cleanup(Failed_SysDev);
		result = PTR_ERR(keyb.sys_dev_ptr);
		goto End;
	}

End:
	pr_devel("keyb4x4_init(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static void __exit keyb4x4_exit(void)
{
	cleanup(ModuleExit);
}

// -----------------------------------------------------------------------------
module_init(keyb4x4_init);
module_exit(keyb4x4_exit);

MODULE_AUTHOR("mBoss <marek.laznicka@elvoris.cz>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Simple but useful device drive skeleton");
MODULE_VERSION("0.0.1");
